/* server.h
 *
 * Data structures, function prototypes and macros used only by the server
 */
#if !defined(SERVER_H)
#define SERVER_H

#include "sim.h"

/* simserver
 * Each server maintains a list of connecting servers for searching
 * when a new player request comes it.
 */
typedef struct simserver {
	char *addr;		/* hostname of server */
	struct in_addr ip;	/* ip address */
	int port;		/* port sim server is listening on */
	struct simserver *next;
	} simserver;

/* map
 * Data structure used by the server to store general info about it's map.
 */
typedef struct map {
	char *server;		/* host name of this server */
	struct in_addr ip;	/* ip address of this server */
	int port;		/* port this server is listening on */
	char *name;		/* name of map */
	char *creator;		/* creator of map */
	char *mapdir;		/* directory map and other files are in */
	char *mapfile;		/* base filename of map */
	char *password;		/* password for transfers */
	char *safenode[10];	/* where to put new players */
	int safenodec;		/* number of safenodes */
	char *expectedhost;	/* address of host this server should be on */
	int   expectedport;	/* port this server should be on */
	simserver *shead;	/* list of connecting servers */
	char *ndfile;		/* file containing nodes */
	char *ptfile;		/* file containing paths */
	char *shfile;		/* shapes */
	char *obfile;		/* objects */
	char *prfile;		/* persons */
	char *plfile;		/* players */
	char *acfile;		/* actions */
	int playerc;		/* currently connected players */
	} map;


/* shapecol
 * An rgb triple defining the ideal colour of a shape
 */
typedef struct shapecol {
	int red, green, blue;	/* range from 0-255 */
	} shapecol;


/* point
 * One position or vector in 3d space
 */
typedef struct point {
	double x,y,z;
	} point;


/* shape
 * The shape struct is used by the server to represent 3d objects in the
 * simulation. Graphical clients do not make use of shapes, instead converting
 * 3d shape descriptions directly to their native representation.
 */
typedef enum { st_poly, st_sphere, st_box, st_cone, st_tube,
	       st_disc, st_coll, st_copy } shape_t;

#define MAX_FACES 50	/* max faces in a polygon shape */
typedef struct shape {
	char name[256];		/* unique identifier */
	struct shape *next;
	struct shape *ne;
	shapecol colour;	/* ideal colour of shape */
	shape_t type;		/* polygon, sphere, etc.. */
	point pos;		/* position points are offset from */
	int hidden;		/* nonzero = in inventory, dont send */
	double rota;		/* angle of rotation in degrees */
	double rotx,roty,rotz;	/* axis to rotate about */
	double scx, scy, scz;	/* scaling */
	int surf;		/* flat, glass, etc.. */
	char *cmt;		/* comment on this shape */
	struct {
		int w,h;	/* size of texture, or 0 if none */
		unsigned char *data;
		} tex;		/* texture info */
	union {
		struct {	/* Polygon - list of points */
			point *pt[MAX_FACES];
			int    ptc[MAX_FACES];
			int facec;
			} poly;
		struct {	/* Sphere - radius and center */
			point cen;
			double rad;
			} sphere;
		struct {	/* Box - two corner points */
			point c1, c2;
			} box;
		struct {	/* Cone - base center, height and radius */
			point base;
			double height;
			double rad;
			} cone;
		struct {	/* Tube - base, height and radius */
			point base;
			double height;
			double rad;
			} tube;
		struct {	/* Collection - a list of shapes */
			struct shape *head;
			} coll;
		struct {	/* Disc - center and radius */
			point cen;
			double rad;
			} disc;
		struct {	/* Copy of another shape */
			struct shape *src;
			} copy;
		} shp;
	} shape;


/* variable
 * Each action has a number of named variables. In addition to declared 
 * variables, there are 4 special variables in each action -
 *  str object - name of this object
 *  str player - name of activating player
 *  str method - name of current method
 *  str params - parameters to current method call
 * This data structure is also used for returning values from expressions in
 * exec.c
 */
#define vt_void		0
#define vt_dbl		1
#define vt_str		2
#define vt_var		3
#define vt_dblarray	4
#define vt_strarray	5
#define vst_static	0
#define vst_local	1
typedef union {
	double dbl;
	char *str;
	struct variable *var;
	struct {
		struct variable *a;	/* block of variables */
		int n;			/* current size */
		int ni;			/* initial size, for locals */
		} arr;
	} value;

typedef struct variable {
	int type;		/* vt_dbl, vt_str, ... */
	char *name;		/* unique to action */
	int storage;		/* vts_local or vts_static */
	struct pnode *init;	/* initialization expression */
	value val;		/* value of the variable */
	struct variable *next;	/* for lists only */
	} variable;


/* pnode
 * A node in the parse tree for a method. Methods are executed by walking
 * the parse tree.
 */
#define pn_if		0	/* statements */
#define pn_block	1
#define pn_cmds		2
#define pn_while	3
#define pn_for		4
#define pn_do		5
#define pn_estat	6
#define pn_null		7
#define pn_equals	20	/* numeric/string expressions */
#define pn_notequals	21
#define pn_add		22	/* numeric expressions */
#define pn_subtract	23
#define pn_mult		24
#define pn_divide	25
#define pn_mod		26
#define pn_greater	27
#define pn_less		28
#define pn_grequals	29
#define pn_lsequals	30
#define pn_or		31
#define pn_and		32
#define pn_var		33
#define pn_assign	34
#define pn_const	35
#define pn_addeq	36	/* op= expressions */
#define pn_subtracteq	37
#define pn_multeq	38
#define pn_divideeq	39
#define pn_uminus	40	/* unary expressions */
#define pn_uplus	41
#define pn_unot		42
#define pn_concat	50	/* string expressions */
#define pn_sconst	51
#define pn_lval		60	/* others */
#define pn_fcall	61
#define pn_preinc	62
#define pn_postinc	63
#define pn_predec	64
#define pn_postdec	65
#define pn_sizeof	66
#define pn_typeof	67
#define pn_aconst	70	/* array expression */
#define pn_aindex	71
typedef struct pnode {
	int type;		/* one of the pn_ things above */
	struct pnode *next;	/* for blocks */
	int lineno;		/* line in source */
	union {
		/* statements - return nothing
		 */
		struct {
			struct pnode *cond;	/* controlling expression */
			struct pnode *tr;	/* true branch */
			struct pnode *fl;	/* false branch */
			} nif;
		struct {
			struct pnode *st;	/* first statement in block */
			} nblock;
		struct {
			struct pnode *cond;	/* controlling expression */
			struct pnode *blk;	/* code to loop through */
			} nwhile;
		struct {
			struct pnode *cond;	/* end of loop expression */
			struct pnode *blk;      /* code to loop through */
			} ndo;
		struct {
			struct pnode *init;	/* initial statement */
			struct pnode *cond;	/* controlling expression */
			struct pnode *blk;	/* code to loop through */
			struct pnode *inc;	/* incremenent statement */
			} nfor;
		struct {
			struct pnode *exp;	/* expression */
			} nestat;
		/* numeric expressions - return a number
		 */
		struct {
			struct pnode *l, *r;
			} nbinop;
		struct {
			struct pnode *var;	/* parse node of lvalue */
			} nvar;
		struct {
			struct pnode *lv;	/* pointer to variable */
			struct pnode *val;	/* value to assign */
			} nassign;
		struct {
			double n;		/* constant value */
			} nconst;
		struct {
			struct pnode *lv;	/* pointer to variable */
			struct pnode *val;	/* value to add/sub/etc.. */
			} nlvalop;
		struct {
			struct pnode *r;	/* val operator is applied to */
			} nunop;
		/* string expressions
		 */
		struct {
			struct pnode *l, *r;
			} nconcat;
		struct {
			char *str;
			} nsconst;
		/* lvalues - return the location of a variable
		 */
		struct {
			variable *loc;		/* pointer to this var */
			} nlval;
		/* Function call
		 */
		struct {
			char *name;		/* function name */
			struct pnode *ar;	/* arg list */
			} nfcall;
		/* Pre/post increment/decrement
		 */
		struct {
			struct pnode *v;	/* variable to change */
			} nincdec;
		/* Size of array
		 */
		struct {
			struct pnode *a;	/* array */
			} nsizeof;
		/* Type of expression
		 */
		struct {
			struct pnode *e;	/* expression */
			} ntypeof;
		/* Array expressions
		 */
		struct {
			struct pnode *l;	/* list of values */
			} naconst;
		struct {
			struct pnode *arr;	/* array */
			struct pnode *ind;	/* index */
			} naindex;
		} val;
	} pnode;


/* method
 * Code called by 'do'ing one of it's names to an object
 */
#define MAX_NAMES 10			/* max names for an action */
#define MAX_MVARS 30			/* max local variables */
typedef struct method {
	char *name[MAX_NAMES];		/* names for this method */
	int namec;			/* number of names */
	variable *var[MAX_MVARS];	/* local varaibles */
	int varc;
	pnode *ptree;			/* parse tree from method code */
	} method;

/* action
 * The action struct determines what an object does then something is done
 * to it. Each action belongs to one object only. The key components of
 * the action struct are variables (kept for the lifetime of the object) and
 * method (called when something is done to the object)
 */
#define MAX_AVARS	30		/* max variables per action */
#define MAX_AMETHODS	20		/* max methods per action */
typedef struct action {
	char name[256];			/* unique name */
	struct action *next;		/* next action in hash table */
	struct action *ne;		/* next action in list */
	variable *var[MAX_AVARS];	/* global variables */
	int varc;
	method *met[MAX_AMETHODS];	/* action methods */
	int metc;
	time_t lastthought;		/* time of last though */
	} action;

/* rfunc
 * Function type for internal function call in executing a method
 */
typedef variable (*rfunc)(action *, pnode *, variable [], int);


/* From netserver.c
 */
void process_connection(int serverfd);
void server_command(int fd, int com, ...);
void server_command_all(int com, ...);
void server_message(int fd, char *fmt, ...);
char *store_to_end(int fd);

/* From shape.c
 */
char *shape_name(void);
shape *create_shape(void);
void add_shape(shape *s);
shape *lookup_shape(char *name);
void reset_shape(void);
shape *next_shape(void);
void remove_shape(shape *s);
void free_shape(shape *sh);
void move_shape(shape *s, double rx, double ry, double rz);
void shade_shape(shape *s, int shd);
shape *copy_shape(shape *s);
shape *deep_copy_shape(shape *s);
void unname_shape(shape *s);
int shape_box(shape *sh, double *x1p, double *y1p, double *z1p,
			 double *x2p, double *y2p, double *z2p);

/* From server.c
 */
void process_request(player *p);
void control_others(void);
void dump_state(void);
int disconnect_players(void);
void delete_player(player *p);
void delete_person(person *pr);
void init_data_dir(char *dir);
void person_move(person *pr, node *nn);
void disconnect_player(player *p, int tellothers);
long object_data_size(object *o);
long player_data_size(player *p);
void check_old_players(void);
char *avatar_sanity_check(shape *sh);

/* From mapparser.y
 */
int parsemap(int filed);
void write_map(int fd, int tl);

/* From shapeparser.y
 */
int parse_shape(int filed, shape **sptr);
int parse_shapes(int filed);
void write_shape(int fd, int tl, shape *sh, int lvl, int toserver);
void write_shapes(int fd, int lvl, int toserver);
void write_all_shapes(int fd);
void write_one_shape(int fd, int tl, shape *sh);

/* From mimeserver.c
 */
char *mime_extension(char *fn);
int send_mime_data(int fd, char *file);
void send_mime_string(int fd, char *dat, char *type);
int send_url_data(int pfd, char *url);

/* From action.c
 */
action *create_action(void);
void add_action(action *a);
action *lookup_action(char *name);
void reset_action(void);
action *next_action(void);
void remove_action(action *a);
void clear_actions(void);
void free_action(action *a);
variable *create_variable(char *name, int type, int storage);
void free_variable(variable *v);
void freev(variable *v);
pnode *create_pnode(void);
void free_pnode(pnode *p);
method *create_method(void);
void free_method(method *m);
variable *lookup_variable(char *name, action *ac, method *mt);

/* From actionparser.y
 */
int parse_action(int filed, action **aptr);
int parse_actions(int filed);
void write_action(int fd, int tl, action *a);
void write_actions(int fd);

/* From exec.c
 */
int exec_action(action *a, player *p, char *m, char **par, int parc);
rfunc find_runtime_func(char *name);
int init_vars(action *a, method *m, int st);

/* globals
 */
extern map *mp;		/* info on this server and it's map */

#endif /* !defined(SERVER_H) */

